import os
# Housekeeping, set num threads to 1 before importing numpy
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
from time import time
import sys
from joblib import Parallel, delayed
from interpolation.splines import eval_linear
import importlib as imp # for reading input files
import h5py

# domestic
import src.handle as hdl
import src.mask as msk
from src.misc import qchi


class data:
    """ A class to save TexTOM data in an object with some information

    Takes input from a designated input file, see input/input_template.py
    for how it has to be stuctured
    """
    def __init__(self, startFileName, single=False, detmsk=None ):
        """Initializes the texture tomography model class

        Calculates arrays that do not need to be updated when fitting/plotting
        Initializes some arrays that can be varied by fitting

        Parameters
        ----------
        startFileName : str
            path to the input file
        
        Attributes created
        ------------
        data_rebinned : 
        """
        ## Load the startfile into variables in the object
        startfile = imp.import_module(startFileName)
        del sys.modules[str(startFileName)] # this is so one can change input files and reload them
        startfile = imp.import_module(startFileName) # else the module stays the same..

        self.title,\
            _, self.q, self.dchi, _, _, _,\
            _,_, self.fov, _,\
            self.shifts, self.datapath,\
            _,_,_,_, \
            self.data_pickle, self.data_load\
        = startfile.input()

        if self.data_load:
            print("Loading data from pickle %s" % self.data_pickle )
            t0=time()
            self = hdl.load_pickle( self.data_pickle, self )

            if single:
                self.data = self.data[0,single].reshape(
                    (1,1,self.data.shape[2]))

            print("\t\tfinished (%2.2f s)" % (time()-t0) )

        else:
            if isinstance(self.shifts, str):
                print('Loading data from files')
                # import from matlab file from SASTT alignment
                kappa, omega, _, _, airscat, _ = hdl.load_mat_proj( self.shifts ) 

                files = hdl.get_filelist(self) # List of files to be loaded

                Nrot = kappa.shape[0] # Number of rotations

                airnorm = airscat.mean()/airscat # to normalize for airscattering

                _,_, self.detShape = qchi(self.q, self.dchi) 

                print('\t Rebinning data\n') # should check if it's not already right size
                t0 = time()
                self.data = np.empty( (Nrot, self.fov[0]*self.fov[1], self.detShape[0]*self.detShape[1] ), np.float64 )
                for g in range(Nrot):
                    path_g = os.path.join( self.datapath, files[g][0] )
                    self.data[g] = airnorm[g] * self.rebin_projection( path_g )
                    # can print counting here
                    sys.stdout.write('\r\t\tProjection %d / %d' % (g+1,Nrot))
                    sys.stdout.flush()
                print('\t finished in %.2f min' % ((time()-t0)/60))
            
            if isinstance(detmsk, str):
                self = hdl.load_pickle( detmsk, self )
            else:
                self.mask_detector = msk.mask_detector( self.data[0,0].reshape(self.detShape) )
                hdl.save_pickle( 'mask_'+self.title, self, 'mask')

            hdl.save_pickle( self.data_pickle, self, 'data' )

    def rebin_projection( self, path ):
        
        with h5py.File(path,'r') as hf:
            q_in = hf['radial_units'][0]
            phi_in = hf['azimuthal_units'][0]*np.pi/180
            all_data_in = hf['cake_integ'][:]

        Q, Phi, _ = qchi(self.q, self.dchi)

        # rebin data to desired q and phi range
        datarange = (phi_in.min(), phi_in.max(), phi_in.shape[0]),\
                    (q_in.min(), q_in.max(), q_in.shape[0])    
        detpoints = np.column_stack( ( Phi-np.pi, Q ) )

        data_rebinned = Parallel(n_jobs=128)(
            delayed(eval_linear)(datarange, data, detpoints) for data in all_data_in)

        return np.array( data_rebinned )