import os
# Housekeeping, set num threads to 1 before importing numpy
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1"
import numpy as np
from time import time
import sys
from typing import Optional
from lmfit import Parameters, Minimizer

# domestic
import src.handle as hdl
import src.mask as msk
from src.misc import sparsemult, integrate_c
from config import n_threads
from src.model import model

from numba import njit, prange, set_num_threads
import warnings
from numba.core.errors import NumbaDeprecationWarning, NumbaPerformanceWarning
warnings.simplefilter('ignore', category=NumbaPerformanceWarning)
set_num_threads(n_threads)

class fitting:
    """ A class to that contains loss function and gradients for fitting
    texture tomography data

    Takes input from a model and a data class object or a previously saved 'fit' pickle
    """
    def __init__(self, dat, mod: Optional[model] = None ):
        """

        Parameters
        ----------
        dat : data class object or str
            if string it needs to be a 'fit'-pickle name
        mod : model class object
            only required if no pickle is loaded
        
        Attributes created
        ------------

        """
        if isinstance(dat, str):
            print("Loading fitting data from pickle %s " % dat )
            t0 = time()
            # in case the name of a pickle is given as input
            self = hdl.load_pickle( dat, self )
            print('\tFinished(%2.2f s)' % (time()-t0) )

        else:
            print('Initialize Fit')
            # import stuff from data and mod objects
            self.data = dat.data
            self.mask_detector = dat.mask_detector
            self.kap_zero = np.where(mod.Kappa == 0)[0]

            # self.scanmask = msk.mask_crazyvoxels(dat)
            self.scanmask = np.ones(
                (self.data.shape[0],self.data.shape[1]),
                bool)

            ## initial guess for c0 from tomogram
            tomogram = mod.tomogram.flatten()/mod.tomogram.max() # normalize so c0 will be around 1
            tomogram[tomogram<0] = 0
            self.C = np.zeros((tomogram.size,1), np.float64)
            self.C[mod.mask_voxels, 0] = tomogram[mod.mask_voxels] # initial guess for c0

            ## renormalize the model
            gt = (0, mod.t0)
            iend = np.searchsorted( mod.iBeams[gt[0],gt[1],:],2**32-1 ) # for sparsearray
            T = sparsemult( 
                mod.Beams[gt[0],gt[1],:iend], 
                mod.iBeams[gt[0],gt[1],:iend], tomogram )
            # the simulated intensity will be multiplied by this to have C~O(1)
            self.sim_fac = self.data[gt].max()/mod.difflets[0][self.mask_detector].max() / T
            self.difflet0_1D = mod.difflets[0].reshape(mod.detShape)[:,0]
            self.set_orders(mod, 0)

            self.q = mod.Qq_det.reshape(mod.detShape)[:,0] # 1D q-vector
            self.qmask = np.ones_like(self.q).astype(bool)
        
        self.choose_projections()

    def mask_q(self, mask=None):
        """ Choose which q-values you want to fit, data will be reduced 
        
        Parameters
        ----------
        Optional:
        mask : 1D ndarray, bool
            needs to be of the length of the q-vector
        """
        if np.any(mask):
            self.qmask = mask
        else:   
            data_powder_1D = _get_powder_1D(self.data, self.mask_detector, 
                            self.detShape, self.iBeams, self.scanmask)
            self.qmask = msk.mask_q_fit( self.q, 
                data_powder_1D, self.difflet0_1D, 'Choose q-ranges to fit')

        qmask2D = np.tile( self.qmask, (self.detShape[1],1) ).T
        self.mask_detector = np.logical_and( 
            qmask2D.flatten(), self.mask_detector )
        self.detShape = (self.qmask.sum(), self.detShape[1])
        self.data = self.data[:,:,self.mask_detector]

    def prepare_fit_c0( self ):
        """ Calculates some properties for fast estimation of the zero order """
        self.total_im = self.data.sum(axis=2) # total intensity for each image
        self.total_proj = self.total_im.sum(axis=1) # total intensity for each projection
        self.data_av = self.total_im / self.difflets_rot_masked[0,0].sum() 
                
    def set_orders( self, mod:model, n_max ):
        """ Changes the HSH order

        Calculates diffractlets and updates the fit object
        using functions in the model class

        Parameters
        ----------
        mod : model class object
        n_max : int
            highest HSH order for fitting
        """
        mod.set_orders( n_max )
        c_tmp = self.C
        self.update_mod(mod)
        if self.C.shape[1] >= c_tmp.shape[1]:
            self.C[:,:c_tmp.shape[1]] = c_tmp
        else:
            self.C = c_tmp[:,:self.C.shape[1]]
        
    def update_mod(self, mod:model):
        """ Imports stuff from the model class

        Parameters
        ----------
        mod : model class object
        """
        self.title, self.Beams, self.iBeams = mod.title, mod.Beams, mod.iBeams
        self.detShape = mod.detShape
        self.difflets_rot_masked = mod.difflets_rot[:,:,self.mask_detector] * self.sim_fac
        self.Sn = mod.Sn
        self.ns = mod.ns
        self.C = np.zeros((mod.x_p.shape[0], mod.difflets.shape[0]))
        print('\tupdated fit object')

    def choose_projections(self, mode='full'):
        """ Decides which sample rotations to use for fitting

        Parameters
        ----------
        mode : str
        """
        # Choose projections:
        Ng0 = self.Beams.shape[0] # total number of rotations
        if mode=='notilt': # only kappa = 0
            self.chosen_proj = self.kap_zero
        elif mode=='half': # every second
            self.chosen_proj = range(0,Ng0,2)
        elif mode=='third': # every third
            self.chosen_proj = range(0,Ng0,3)
        elif mode=='asnotilt': # equidistant, as many as at kappa = 0
            n = self.kap_zero.size
            self.chosen_proj = np.arange(0,Ng0,int(Ng0/n))[:n]
        else: # all of them
            self.chosen_proj = range(Ng0)
        print('Choose projections: %s' % mode)

    def loss( self, C ):
        """ Wrapper for a numba-compiled function calculating the loss function

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        loss : float
        """
        loss = _loss( C,
            self.data[self.chosen_proj], self.scanmask[self.chosen_proj],
            self.Beams[self.chosen_proj], self.iBeams[self.chosen_proj], 
            self.difflets_rot_masked[self.chosen_proj],
            )
        return loss

    def loss_c0( self, C ):
        """ Wrapper for a numba-compiled function calculating the loss function
        using only the zero-order coefficients compared to the averaged images

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        loss : float
        """
        loss = _loss_c0( C,
            self.data_av[self.chosen_proj], self.scanmask[self.chosen_proj],
            self.Beams[self.chosen_proj], self.iBeams[self.chosen_proj],
            )
        return loss
    
    def grad_full( self, C ):
        """ Wrapper for a numba-compiled function calculating the gradient
        of the loss function for all sHSH coefficients

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        grad : 2D ndarray, float
            gradient for all the sHSH coefficients
            dim: 0: voxel index, 1: sHSH index
        """
        grad = _grad_full( C,
            self.data[self.chosen_proj], self.scanmask[self.chosen_proj],
            self.Beams[self.chosen_proj], self.iBeams[self.chosen_proj], 
            self.difflets_rot_masked[self.chosen_proj],
            )
        return grad

    def grad_highest( self, C ):
        """ Wrapper for a numba-compiled function calculating the gradient
        of the loss function for the highest order sHSH coefficients 

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        grad : 2D ndarray, float
            gradient for all the sHSH coefficients
            dim: 0: voxel index, 1: sHSH index
        """            
        grad = _grad_highest( C,self.Sn[-1],
            self.data[self.chosen_proj], self.scanmask[self.chosen_proj],
            self.Beams[self.chosen_proj], self.iBeams[self.chosen_proj], 
            self.difflets_rot_masked[self.chosen_proj],
            )
        return grad
    
    def grad_c0( self, C ):
        """ Wrapper for a numba-compiled function calculating the gradient
        of the loss function for the zero order sHSH coefficients from 
        the average image value

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        grad : 2D ndarray, float
            gradient for all the sHSH coefficients
            dim: 0: voxel index, 1: sHSH index
        """
        grad = _grad_c0( C,
            self.data_av[self.chosen_proj], self.scanmask[self.chosen_proj],
            self.Beams[self.chosen_proj], self.iBeams[self.chosen_proj], 
            )
        return grad

    def lipschitz( self ):
        """ Wrapper for a numba-compiled function calculating the Lipschitz 
        constant for the zero order sHSH coefficients

        Return values
        ------------
        Lipschitz : 1D ndarray, float
        """
        Lipschitz = _lipschitz_c0( self.C,
            self.scanmask[self.chosen_proj],
            self.Beams[self.chosen_proj], self.iBeams[self.chosen_proj],
            )
        return Lipschitz   
    
    def projection( self, g, C ):
        """ Wrapper for a numba-compiled function calculating a series
        of diffraction pattern corresponding to one sample orientation

        Parameters
        ----------
        g : int
            rotation index
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index

        Return values
        ------------
        images : 2D ndarray, float
            dim 0: translation, 1: detector points
        """
        print('\tCalculating projection %d' % g )
        images = _projection(
            C, g, self.scanmask,
            self.Beams, self.iBeams, 
            self.difflets_rot_masked,
        )
        return images

    def image( self, C, g, t ):
        """Simulates a single diffraction image for rotation g and 
        translation t

        Parameters
        ----------
        C : 2D ndarray, float
            set of sHSH coefficients for each voxel
            dim: 0: voxel index, 1: sHSH index
        g : int
            rotation index
        t : int
            translation index
        Return values
        ------------
        image : 1D ndarray, float
        """
        iend = np.searchsorted(self.iBeams[g,t,:],2**32-1) # for sparsearray
        # get projected coefficients for the corresponding beam rotation/translation
        c_proj = integrate_c( self.Beams[g,t,:iend], self.iBeams[g,t,:iend], C )
        # calculate the image
        image = c_proj @ self.difflets_rot_masked[g] 
        return image

    def MDL( self, loss ):
        """ Calculates the Medium description length for order selection
        
        M. Hansen and B. Yu, ?Model selection and the principle of minimum 
        description length, Journal of the American Statistical Association,
        vol. 96, no. 454, pp. 746-774, 2001.

        Parameters
        ----------
        loss : float

        Return values
        ------------
        MDL : float
        """
        # Calculate the number of datapoints
        lowmask = np.array(
            [beam[0]<2**23-1 for beam_g in self.iBeams[self.chosen_proj] for beam in beam_g])
        im_mask = np.logical_and( lowmask, self.scanmask[self.chosen_proj].flatten() )
        N_images = np.sum( im_mask )
        N_data = N_images * self.mask_detector.sum()
        # the number of parameters
        n_par = self.C[self.C[:,0]>0].size
        # calculate the medium description length
        MDL = N_data/2 * np.log(loss) + n_par/2 * np.log(N_data)
        return MDL

@njit(parallel=True)
def _loss( C, data, scanmask, Beams, iBeams, difflets_rot ): 
    """ Calculates the loss function as the sum of squared residuals

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data : 3D ndarray, float
        dim: 0: rotation, 1: translation, 2: detector points
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points

    returns: 
    ------------
    loss : float
    """
    loss = 0.0
    for g in prange(Beams.shape[0]): # rotations
        for t in range(Beams.shape[1]): # translations
            if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
                iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
                # get projected coefficients for the corresponding beam rotation/translation
                c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
                # calculate the image
                I_sim = c_proj @ difflets_rot[g] 
                # calculate the squared sum of residuals
                residuals = data[g,t] - I_sim 
                loss += (residuals**2).sum() / (data[g,t]**2).sum()
    return loss

@njit(parallel=True)
def _loss_c0( C, data_av, scanmask, Beams, iBeams ): 
    """ Calculates the loss function as the sum of squared residuals
    using only the zero-order coefficients compared to the averaged images

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data_av : 2D ndarray, float
        dim: 0: rotation, 1: translation
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index

    returns: 
    ------------
    loss : float
    """
    loss = 0.0
    for g in prange(Beams.shape[0]): # rotations
        for t in range(Beams.shape[1]): # translations
            if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
                iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
                # get projected coefficients for the corresponding beam rotation/translation
                c0_proj = sparsemult( Beams[g,t,:iend], iBeams[g,t,:iend], C[:,0] )
                loss += (data_av[g,t] - c0_proj)**2 / (data_av[g,t])**2 
    return loss

@njit(parallel=True)
def _grad_full( C, data, scanmask, Beams, iBeams, difflets_rot ): 
    """ Calculates the gradient for all coefficients

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data : 3D ndarray, float
        dim: 0: rotation, 1: translation, 2: detector points
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points

    Return values
    ------------
    grad : 2D ndarray, float
        gradient for all the sHSH coefficients
        dim: 0: voxel index, 1: sHSH index
    """
    grad = np.zeros_like( C )
    for g in prange(Beams.shape[0]): # rotations
        for t in range(Beams.shape[1]): # translations
            if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
                iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
                # get projected coefficients for the corresponding beam rotation/translation
                c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
                # calculate the image
                I_sim = c_proj @ difflets_rot[g] 
                # get the gradient on the coefficients from this projection
                residuals = data[g,t] - I_sim 
                grad_proj =  - 2 * difflets_rot[g] @ residuals
                # put the beam-intensity weighted value into each voxel
                for isp in range(iend): # sparse index
                    iv = iBeams[g,t,isp] # voxel index
                    grad[iv] += Beams[g,t,isp] * grad_proj
    return grad

@njit(parallel=True)
def _grad_highest(C, order_slice, data, scanmask, Beams, iBeams, difflets_rot ): 
    """ Calculates the gradient for the highest order coefficients

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data : 3D ndarray, float
        dim: 0: rotation, 1: translation, 2: detector points
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points

    Return values
    ------------
    grad : 2D ndarray, float
        gradient for all the sHSH coefficients
        dim: 0: voxel index, 1: sHSH index
    """
    grad = np.zeros_like( C )
    for g in prange(Beams.shape[0]): # rotations
        for t in range(Beams.shape[1]): # translations
            if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
                iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
                # get projected coefficients for the corresponding beam rotation/translation
                c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
                # calculate the image
                I_sim = c_proj @ difflets_rot[g] 
                # get the gradient on the coefficients from this projection
                residuals = data[g,t] - I_sim 
                grad_proj =  - 2 * difflets_rot[g, order_slice[0]:order_slice[1]] @ residuals
                # put the beam-intensity weighted value into each voxel
                for isp in range(iend): # sparse index
                    iv = iBeams[g,t,isp] # voxel index
                    grad[iv, order_slice[0]:order_slice[1]] += Beams[g,t,isp] * grad_proj
    return grad

@njit(parallel=True)
def _grad_c0( C, data_av, scanmask, Beams, iBeams ): 
    """ Calculates the gradient for the zero order coefficients
    from the averaged images

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    data_av : 2D ndarray, float
        dim: 0: rotation, 1: translation
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index

    returns: 
    ------------
    grad : 2D ndarray, float
        gradient for all the sHSH coefficients
        dim: 0: voxel index, 1: sHSH index
    """
    grad = np.zeros_like( C )
    for g in prange(Beams.shape[0]): # rotations
        for t in range(Beams.shape[1]): # translations
            if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
                iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
                # get projected coefficients for the corresponding beam rotation/translation
                c0_proj = sparsemult( Beams[g,t,:iend], iBeams[g,t,:iend], C[:,0] )
                residual = data_av[g,t] - c0_proj
                # assign gradient to the affected voxels
                grad[ iBeams[g,t,:iend], 0 ] += - 2 * residual * Beams[g,t,:iend]
    return grad

@njit(parallel=True)
def _lipschitz_c0( C, scanmask, Beams, iBeams ): 
    """ Calculates the Lipschitz constant for the zero order coefficients

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index

    returns: 
    ------------
    L : 1D ndarray, float
    """
    L = np.zeros_like( C[:,0], np.float64)
    for g in prange(Beams.shape[0]): # rotations
        for t in range(Beams.shape[1]): # translations
            if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
                iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
                # assign contribution to the affected voxels
                L[ iBeams[g,t,:iend] ] += Beams[g,t,:iend]**2
    return L

@njit(parallel=True)
def _projection( C, g, scanmask, Beams, iBeams, difflets_rot ):
    """Simulates diffraction images from a 2D scan over the sample

    Parameters
    ----------
    C : 2D ndarray, float
        set of sHSH coefficients for each voxel
        dim: 0: voxel index, 1: sHSH index
    g : int
        rotation index
    scanmask : 2D ndarray, bool
        dim: 0: rotation, 1: translation
    Beams : 3D ndarray, float
        sparse-array containing beam intensity for all configurations
        dim: 0: rotation, 1: translation, 2: voxel (sparse)
    iBeams : 3D ndarray, int
        array of indices for the sparse Beams array
        dim: 0: rotation, 1: translation, 2: voxel index
    difflets_rot : 3D ndarray, float
        diffractlets for all sample rotations
        dim: 0: sample rotation, 1: sHSH indices, 2: detector points
        
    Return values
    ------------
    images : 3D ndarray, float
        array of resulting scattering intensity for each point on the 
        detector, for each rotation and translation
        dimensions: 0: rotation, 1: translation, 2: detector points
    """
    images = np.zeros( (Beams.shape[1], difflets_rot.shape[2]), np.float64 )
    for t in range(Beams.shape[1]): # translations
        if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
            iend = np.searchsorted(iBeams[g,t,:],2**32-1) # for sparsearray
            # get projected coefficients for the corresponding beam rotation/translation
            c_proj = integrate_c( Beams[g,t,:iend], iBeams[g,t,:iend], C )
            # calculate the image
            images[t,:] = c_proj @ difflets_rot[g] 
    return images

@njit(parallel=True)
def _get_powder_1D( data, mask_detector, detShape, iBeams, scanmask ):
    data_1D = np.zeros((detShape[0]),np.float64)
    for g in prange(iBeams.shape[0]): # rotations
        for t in range(iBeams.shape[1]): # translations
            if ( scanmask[g,t] and (iBeams[g,t,0] < 2**32-1) ):
                data_1D += _masked_to_1D(
                    data[g,t,mask_detector], mask_detector, detShape)
    return data_1D

@njit
def _masked_to_1D( A, mask_detector, detShape ):
    A_full = np.zeros( detShape, np.float64 ).flatten()
    A_full[mask_detector] = A
    A_1D = np.array([dq[dq>0].mean() for dq in A_full.reshape(detShape)])
    return A_1D
