import numpy as np
import glob
import importlib as imp
import h5py
import pickle
import os
from time import time 
import sys
from interpolation.splines import eval_linear
import shutil
import pyevtk

from config import pickle_path, backup_path
# from src.data import data
# from src.model import model

''' Library for handling input and output'''
"""
Handle Pickles
"""

def save_pickle( name, object, type ):
    ''' Saves an object into the pickles folder
    
    Parameters
    ----------
    name : str
        pickle name without extension
    object : class object
        contains the attributes to be saved
    type : str
        must be 'diff', 'tomo', 'data', 'fit', 'mask' or 'opt'
        saves attributes accordingly
    '''
    t0=time()
    match type:
        case 'diff':
            save_attributes = {
                'symmetry',
                'Gc', 'dV', 'V_fz',
                'detShape', 'Chi_det', 'Qq_det',
                'Isc', 'difflets',
            }
        case 'tomo':
            save_attributes = {
                'Omega', 'Kappa', 'Gs',
                'tomogram', 
                'Ty', 'Tz', 't0', 'fov', 
                'x_p', 'p0', 'nVox', 
                'mask_voxels', 'cutoff_low',
                'Beams', 'iBeams', 
            }
        case 'data':
            save_attributes = {
                'data_rebinned',
                'mask_detector',
            }
        case 'fit':
            save_attributes = {
                'title',
                'C', 'sim_fac', 'Sn',
                'q', 'qmask',
                'difflet0_1D',
                'data', 
                'kap_zero', 'detShape',
                'scanmask', 'mask_detector',
                'iBeams', 'Beams'
            }
        case 'mask':
            save_attributes = {
                'mask_detector'
            }
        case 'opt':
            save_attributes = {
                'loss', 'nGradA', 'nGradC',
                'projA', 'projC', 'fgam',
                't_iter_av', 't_total',
                'max_memory'
            }
    obj = {}
    for attr in save_attributes:
        obj[attr] = getattr(object, attr)
    os.makedirs(pickle_path, exist_ok=True) 
    fid = open( os.path.join(pickle_path,name+'.txt'), 'wb' )
    pickle.dump(obj, fid)
    fid.close()
    print('\tSaved object to pickle %s (%d s)' % (name, int(time()-t0)) )

def load_pickle( name, object ):
    ''' Adds attributes found in pickle to the object
    
    Parameters
    ----------
    name : str
        pickle name without extension
    object : class object
        contains the attributes to be saved

    Return values
    ------------
    object : modified class object
    '''
    fid = open( os.path.join(pickle_path,name+'.txt'), 'rb' )
    pickobj = pickle.load( fid )
    fid.close()

    for key in pickobj.keys():
        setattr( object, key, pickobj[key] )
    
    return object

def load_opt( path ):
    ''' Loads a pickle to a dictionary
    
    Parameters
    ----------
    path : str
        full path to the pickle

    Return values
    ------------
    picdic : dict
    '''
    fid = open( path, 'rb' )
    picdic = pickle.load( fid )
    fid.close()
    return picdic

def backup_pickles( path=None ):
    ''' Copies everything from the pickles folder to path
    
    Parameters
    ----------
    path : str
        backup directory, if None, takes the path from config.py
    '''
    if path==None and backup_path:
        os.makedirs(backup_path, exist_ok=True) 
        copy_folder_contents(pickle_path, backup_path)

def copy_folder_contents(source_folder, destination_folder):
    ''' copies everything from source to destination '''
    # Ensure the destination folder exists
    if not os.path.exists(destination_folder):
        os.makedirs(destination_folder)

    # Iterate over all files and subdirectories in the source folder
    for item in os.listdir(source_folder):
        source_item = os.path.join(source_folder, item)
        destination_item = os.path.join(destination_folder, item)

        # Check if the item is a file
        if os.path.isfile(source_item):
            # Skip if the file already exists in the destination folder
            if not os.path.exists(destination_item):
                shutil.copy2(source_item, destination_item)
                print(f"\tCopied: {item}")
            else:
                print(f"\tSkipped (File already exists): {item}")

        # Check if the item is a subdirectory
        elif os.path.isdir(source_item):
            # Recursively copy the subdirectory
            copy_folder_contents(source_item, destination_item)

""" 
Handle internal files
"""
def save_projections(images, rotation, path):
    dirpath = path.split('/')[0:-1]
    dir = '/'.join(dirpath)
    if os.path.isdir(dir) == False:
        os.mkdir(dir)
        print('Created path')
    h5id = h5py.File(path,'w')
    for g in range(images.shape[0]):
        dset = h5id.create_dataset( "proj"+str(g).zfill(3), images[g,:,:].shape, 'f')
        dset[:,:] = images[g,:,:]
        dset.attrs['rotation'] = rotation[g]
    print('saved projections to file: ' + path)

def load_projections(path, shape):
    images = np.empty(shape)
    rot = []
    with h5py.File(path,'r') as hf:
        sets = list(hf.keys())
        for g in range(shape[0]):
            for t in range(shape[1]):
                images[g,t,:] = hf[sets[g]][t]
            rot.append( hf[sets[g]].attrs['rotation'] )
    print('loaded data from file: ' + path)
    return images, rot

def save_coefficients( C, mod, path = 'output/sample.txt' ):
    ## writes into output/sample.txt
    # one line per voxel:
    # index, coordinates, overall scaling A, isotropic part U, sHSH coefficients cs
    px = 1 # spatial precision
    pc = 6 # coefficient precision

    os.makedirs('output', exist_ok=True) 
    fid  = open( path, 'w') 
    fid.write('i\t x\t y\t z\t c\n')
    for ivox in range( mod.x_p.shape[0] ):
        fid.write('{:.0f}\t{:.{px}f}\t{:.{px}f}\t{:.{px}f}'.format(
            ivox, *mod.x_p[ivox], px=px, pc=pc))
        for c in C[ivox]:
            fid.write('\t{:.{pc}f}'.format(
                c, pc=pc)) 
        fid.write('\n')
    fid.close()
    print('saved parameters to file: ' + path)

def load_coefficients( path ):
    t0 = time()
    data = np.genfromtxt(path, skip_header=1)
    C = data[:,4:]
    print('loaded parameters from file (%.2f s): %s' % (time()-t0, path)) 
    return C

"""
Functions handling X-ray data
"""

def load_data( path, n ):
    with h5py.File(path,'r') as hf:
        q = hf['radial_units'][0]
        phi = hf['azimuthal_units'][0]
        data = hf['cake_integ'][n]
    return q, phi, data

def load_data_rebin( mod, path, n):
    with h5py.File(path,'r') as hf:
        q = hf['radial_units'][0]
        phi = hf['azimuthal_units'][0]*np.pi/180
        data = hf['cake_integ'][n]

    # rebin data to desired q and phi range
    datarange = (phi.min(), phi.max(), phi.shape[0]),\
                (q.min(), q.max(), q.shape[0])    
    detpoints = np.column_stack( ( mod.Phi_ew-np.pi, mod.Qq_ew ) )
    data_rebinned = eval_linear( datarange, data, detpoints )

    return data_rebinned

def load_mat_proj( path ):
    ''' Loads a file from SASTT alignment
    
    Parameters
    ----------
    path : str

    Return values
    ------------
    kappa, omega: 1D ndarray, float
        tilt and rotation angle
    shiftx, shiftz : 1D ndarray, float
        shifts of projections for alignment
    air_scat : 1D ndarray, float
        intensity correction from air scattering
    tomogram : 1D ndarray, float
        resulting tomogram from SASTT
    '''
    f = h5py.File(path,'r')
    n_proj = f['projection/dx'][()].shape[0]
    kappa = np.array([ f[f.get('projection/rot_x')[k][0]][0][0] for k in range(n_proj) ])
    omega = np.array([ f[f.get('projection/rot_y')[k][0]][0][0] for k in range(n_proj) ])
    shiftx = np.array([ f[f.get('projection/dx')[k][0]][0][0] for k in range(n_proj) ])
    shiftz = np.array([ f[f.get('projection/dy')[k][0]][0][0] for k in range(n_proj) ])
    # filename = np.array([ f[f.get('projection/todo_list')[k][0]][0][0] for k in range(n_proj) ])
    # pad = np.array([ f[f.get('projection/pad')[k][0]][0][0] for k in range(n_proj) ])
    air_scat = np.array([ f[f.get('projection/air_scattering')[k][0]][0][0] for k in range(n_proj) ])
    tomogram = np.array(f.get('tomogram'))
    return kappa, omega, shiftx, shiftz, air_scat, tomogram

def get_filelist( dat ):
    kappa, omega, _, _, _, _ = load_mat_proj(dat.shifts)
    kapstr = [('%d'%k).replace('-','m') for k in kappa]
    omestr = [('%08.2f'%o).replace('.','p') for o in omega]
    nameref = ['_%s_%s_'%(k,o) for k,o in zip(kapstr,omestr)]

    filenames = np.array( next(os.walk(dat.datapath), (None, None, []))[2] ) # [] if no file

    filelist = []
    missing = []
    duplicates = []
    for nam in nameref:
        match = [str(s) for s in filenames if nam in s]
        # match = next((str(s) for s in filenames if nam in s), None) # this is faster but only gives the first find!
        if (len(match) == 0):
            missing.append(nam)
        elif (len(match) >1 ):
            duplicates.append(match)
        else:
            filelist.append(match)

    print('\t Loaded filelist and compared with input')
    print('\t\t missing: %d' % len(missing) )
    [print(miss) for miss in missing]
    print('\t\t duplicates: %d' % len(duplicates) )
    [print(dupl) for dupl in duplicates]

    # this moves duplicates to a different folder
    # path_dupl = os.path.normpath(path)+'_duplicatesele/'
    # os.makedirs( path_dupl, exist_ok=True)
    # for dupl in duplicates:
    #     os.rename(path+dupl[1], path_dupl+dupl[1])

    return filelist

def get_list_opt( path ):

    lst = []
    with open(path, 'r') as fid:
        lines = fid.readlines()    

    for line in lines:
        file = line.strip()
        i0 = file.find('nmax')
        i1 = file.find('_', i0)
        nmax = int( file[i0+4 : i1] )
        lst.append([file, nmax])
    return(lst)

def make_list_proj( basedir, listname, exclude):
    """
    Example input
    basedir = '/dataSATA/data_synchrotron/esrf_id13_2023_09_ihmi1513_biomorphs/ihmi1531/id13/20230914/RAW_DATA/helix_s7/'
    listname = 'data/list_helix_s7.txt'
    exclude = ['_a_','_b_','_c_',]
    """

    dirs = [x[0].split('/')[-1] for x in os.walk(basedir)]
    dirs_data, kappa, omega = [], [], []
    omega = []
    for dir in dirs:
        ex = [k in dir for k in exclude]
        if any(ex):
            continue
        if dir[-4:]=='diff':
            angles_str = dir.split('_')[-3:-1]
            kap = float(angles_str[0].replace('m','-'))
            ome = float(angles_str[1].replace('p','.'))


            kappa.append(kap)
            omega.append(ome)
            dirs_data.append(dir)
    kappa, omega = np.array(kappa), np.array(omega)

    fid  = open( listname, 'w') 
    fid.write(basedir)
    fid.write('\n kap\t ome\t file\t duplicate')

    for kap in np.unique(kappa)[::-1]:
        ord_kap = np.where(kappa==kap)[0].tolist()
        ome_sub = omega[ ord_kap ]
        ord_ome = np.argsort( ome_sub ).tolist()
        files_sub = [dirs_data[i] for i in ord_kap]
        files_ord = [files_sub[i] for i in ord_ome]
        ome_ord = ome_sub[ord_ome]
        bool_dup = ome_ord[:-1]-ome_ord[1:]
        # bool_dup = np.logical_not( np.concatenate([bool_dup,[0,]]), np.concatenate([[0,],bool_dup]) )
        dup = [ '' if i else '*' for i in bool_dup ]
        dup.insert(0,'')

        for k in range(ome_ord.shape[0]):
            fid.write('\n %u \t %5.2f \t %s \t %s' %( kap, ome_ord[k], files_ord[k], dup[k] ))
        
    fid.close()

def load_list_proj( path ):

    kappa, omega, filenames = [], [], []

    with open(path, 'r') as f:
        content = f.readlines()

        path = content[0].strip()
        for line in content[2:]:
            values = line.strip().split()
            kappa.append(float(values[0]))
            omega.append(float(values[1]))
            filenames.append(values[2])
    
    return kappa, omega, filenames, path
            

"""
Handle output
"""
def write_vtk_direct( mod, path, ivox, scalars=[], vectors=[], notes='notes' ):
        Nvox = ivox.shape[0]

        print('Writing Paraview file to path:\n\t%s' % path)
        t0 = time()
        with open(path, 'w') as file:
            # Write VTK header (256 char maximum)
            file.write("# vtk DataFile Version 3.0\n")
            file.write("%s\n" % notes[:256]) # header, 256 char maximum
            file.write("ASCII\n")
            file.write("DATASET UNSTRUCTURED_GRID\n")

            # Write points (voxel positions)
            file.write("POINTS %d float\n" % Nvox)
            for p in mod.x_p[ivox]:
                file.write("{} {} {}\n".format(*p))

            file.write("POINT_DATA %d\n" % Nvox)

            for scal in scalars:
                file.write("SCALARS %s float\n" % scal[0])
                file.write("LOOKUP_TABLE default\n")
                for v in scal[1]:
                    file.write("{}\n".format(v))

            for vec in vectors:
                file.write("VECTORS %s float\n" % vec[0])
                for ax in vec[1]:
                    file.write("{} {} {}\n".format(*ax))

def write_vtk( mod, A, c, path, notes='TexTOM sample data', tresh=0., SASTT=True ):

    # get chosen voxels to show
    ishow, a_pref, b_pref, c_pref, odf_max, \
        a_mean, b_mean, c_mean, var, \
            odf_norm, odf_neg = mod.g_pref_all(
                                    A, c, tresh, mean=False) 
    Nvox = ishow.shape[0]

    if SASTT:
        Nvox_tot = mod.nVox[0]*mod.nVox[1]*mod.nVox[2]

        # A_w = np.flip(np.flip(A.reshape(mod.nVox).transpose((0,2,1)),2),0)
        A_w = A.reshape(mod.nVox).transpose((1,2,0))#
        A_w = np.ascontiguousarray(A_w)

        odf_max_full = np.zeros( (Nvox_tot), np.float64)
        odf_max_full[ishow] = odf_max
        odf_max_full = odf_max_full.reshape( mod.nVox )
        c_pref_full = np.zeros( (Nvox_tot,3), np.float64)
        c_pref_full[ishow] = c_pref
        c_pref_full = c_pref_full.reshape( (*mod.nVox,3) )
        # c_pref_full = np.array((c_pref_full[:,:,:,0],c_pref_full[:,:,:,1],c_pref_full[:,:,:,2]))
        c_0 = np.ascontiguousarray(
            # np.flip(np.flip(
                c_pref_full[:,:,:,0].transpose((1,2,0)) 
                # ,2),0)
        )
        c_1 = np.ascontiguousarray(
            # np.flip(np.flip(
                c_pref_full[:,:,:,1].transpose((1,2,0)) 
                # ,2),0)
        )
        c_2 = np.ascontiguousarray(
            # np.flip(np.flip(
                c_pref_full[:,:,:,2].transpose((1,2,0)) 
                # ,2),0)
        )

        # high level pyevtk
        pyevtk.hl.imageToVTK(
            path,
            origin=(0,0,0), 
            # spacing=(mod.Dstep, mod.Dstep, mod.Dstep),
            pointData= {
                "A": A_w,
                "c_pref": (c_0, c_1, c_2),
                },
        )

    else:
        print('Writing Paraview file to path:\n\t%s' % path)
        t0 = time()
        with open(path+'.vtk', 'w') as file:
            # Write VTK header (256 char maximum)
            file.write("# vtk DataFile Version 3.0\n")
            file.write("%s\n" % notes[:256]) # header, 256 char maximum
            file.write("ASCII\n")
            file.write("DATASET UNSTRUCTURED_GRID\n")

            # Write points (voxel positions)
            file.write("POINTS %d float\n" % Nvox)
            for p in mod.x_p[ishow]:
                file.write("{} {} {}\n".format(*p))

            file.write("POINT_DATA %d\n" % Nvox)

            # Voxel intensity A
            file.write("SCALARS i_vox int\n")
            file.write("LOOKUP_TABLE default\n")
            for v in ishow:
                file.write("{}\n".format(v))

            # Voxel intensity A
            file.write("SCALARS A float\n")
            file.write("LOOKUP_TABLE default\n")
            for v in A[ishow]:
                file.write("{}\n".format(v))

            # odf max
            file.write("SCALARS odf_max float\n")
            file.write("LOOKUP_TABLE default\n")
            for v in odf_max:
                file.write("{}\n".format(v))

            # # spherical variance after Marcia
            # file.write("SCALARS var float\n")
            # file.write("LOOKUP_TABLE default\n")
            # for v in var:
            #     file.write("{}\n".format(v))

            # odf norm
            file.write("SCALARS odf_norm float\n")
            file.write("LOOKUP_TABLE default\n")
            for v in odf_norm:
                file.write("{}\n".format(v))

            # odf neg
            file.write("SCALARS odf_neg float\n")
            file.write("LOOKUP_TABLE default\n")
            for v in odf_neg:
                file.write("{}\n".format(v))

            # preferred orientation of the crystal a-axis
            file.write("VECTORS a_axis float\n")
            for ax in a_pref:
                file.write("{} {} {}\n".format(*ax))

            # preferred orientation of the crystal b-axis
            file.write("VECTORS b_axis float\n")
            for ax in b_pref:
                file.write("{} {} {}\n".format(*ax))

            # preferred orientation of the crystal c-axis
            file.write("VECTORS c_axis float\n")
            for ax in c_pref:
                file.write("{} {} {}\n".format(*ax))

            # # mean orientation of the crystal a-axis
            # file.write("VECTORS a_axis_mean float\n")
            # for ax in a_mean:
            #     file.write("{} {} {}\n".format(*ax))

            # # mean orientation of the crystal b-axis
            # file.write("VECTORS b_axis_mean float\n")
            # for ax in b_mean:
            #     file.write("{} {} {}\n".format(*ax))

            # # mean orientation of the crystal c-axis
            # file.write("VECTORS c_axis_mean float\n")
            # for ax in c_mean:
            #     file.write("{} {} {}\n".format(*ax))

    # print( '\t finished in %.2f s' % (time()-t0) )

def read_vtk( path ):
    scalars = {}
    vectors = {}
    with open(path) as fid:
        lines = fid.readlines()
        iscl = []
        nscl = []
        ivec = []
        nvec = []
        for k, line in enumerate(lines):
            if line.find('SCALARS') > -1:
                iscl.append(k)
                l = line.split(' ')
                nscl.append(l[1])
            elif line.find('VECTORS') > -1:
                ivec.append(k)
                l = line.split(' ')
                nvec.append(l[1])

    for k in range(len(iscl)):
        if k < len(iscl)-1:
            end = iscl[k+1]
        else:
            end = ivec[0]
        scalars[nscl[k]] = np.array(lines[iscl[k]+2:end], np.float64)
    for k in range(len(ivec)):
        if k < len(ivec)-1:
            end = ivec[k+1]
        else:
            end = len(lines)
        vectors[nvec[k]] = np.array( 
            [line.split(' ') for line in lines[ivec[k]+1:end]], 
            np.float64)
    return scalars, vectors

