import numpy as np
import os
from time import time
from datetime import datetime
import sys
from numba import njit
from resource import getrusage, RUSAGE_SELF
from pickle import dump

from src.fitting import fitting

def optimize( fit:fitting, mode, name='',
              tol=1e-3, minstep=1e-3, itermax=3000):
    """ Gradient descent optimizer for texture tomography

    Parameters
    ----------
    fit : class object
        object from the fitting module containing loss function and gradient
    mode : int
        decides what to optimize
        0: only zero order, 1: only highest order, 2: all orders
    -- optional --
    name : str
        add sth to the name of the output file
    tol : float
        stops optimization if the relative change of the norm of the
        gradient deceeds this value
    minstep : float
        stops optimization if the initial stepsize is reduced by this factor
    itermas : int
        stops optimization if this number of iterations is reached
    
    Return values
    ------------
    c : 1D ndarray, float
        solution found by the optimization in terms of sHSH coefficients
    opt : dict
        collection of results:
        'c' : 1D ndarray, float
            solution found by the optimization in terms of sHSH coefficients
        'loss' : 1D ndarray, float
            value of the loss function at each iteration
        'nGrad' : 1D ndarray, float
            norm of the gradient at each iteration
        'proj' : 1D ndarray, bool
            marks if c0 was projected or not for each iteration
        'prec' : float
            final precision
        'fgam' : float
            final stepsize modification
        't_total' : float
            total optimization time in s
        't_iter_av' : float
            average time per iteration in s
        'max_memory' : float
            memory usage of the process
    """

    t0 = time()
    c = fit.C
    match mode:
        case 0: # fit c0 from average intensity
            print('Start optimizing c0 from average intensity')
            modename = 'c0'
            fit.prepare_fit_c0()
            def fun( C ):
                loss = fit.loss_c0( C )
                return loss
            def jac( C ):
                Jac = fit.grad_c0( C )
                return Jac
            L = fit.lipschitz().reshape((c.shape[0],1))
            L[ L < 1e-4*L.max() ] = 0.
            invL = np.zeros_like(L)
            invL[L!=0] = 2/L[L!=0]
            grad = jac( c )
            gam = 1/np.abs( grad ).max() * invL/invL.max()

        case 1: # fit only highest order
            print('Start optimizing order %d' % fit.ns[-1])
            modename = 'hc'
            def fun( C ):
                loss = fit.loss( C )
                return loss
            def jac( C ):
                Jac = fit.grad_highest( C )
                return Jac
            L = fit.lipschitz()
            L[ L < 1e-4*L.max() ] = 0.
            invL = np.zeros_like(L)
            invL[L!=0] = 2/L[L!=0]
            grad = jac( c )
            gam = 1/np.abs( grad ).max() * np.tile( c[:,0]*invL/invL.max(), 
                                                   (c.shape[1],1)).T 

        case 2: # fit all orders
            print('Start optimizing all orders')
            modename='C'
            def fun( C ):
                loss = fit.loss( C )
                return loss
            def jac( C ):
                Jac = fit.grad_full( C )
                return Jac
            L = fit.lipschitz()
            L[ L < 1e-4*L.max() ] = 0.
            invL = np.zeros_like(L)
            invL[L!=0] = 2/L[L!=0]
            grad = jac( c )
            gam = 1/np.abs( grad ).max() * np.tile( c[:,0]*invL/invL.max(), 
                                                   (c.shape[1],1)).T 

    # Initialize functions    
    loss = fun( c )
    grad = jac( c )
    t1 = time()
    print('\tCompiled functions, %.3f s' % ((t1-t0)) )
    
    # initialize iterables
    c_next = c
    fgam, prec = 1,1
    nGrad = _vnorm(grad)
    c0_proj = []
    opt = {}
    opt['loss'] = [loss]
    opt['nGrad'] = [nGrad]
    opt['proj'] = []
    opt['max_memory'] = getrusage(RUSAGE_SELF)
    iter = 0
    t_iter_av = 0.

    try:
        # start optimization
        while ( 
            ( prec > tol ) and 
            ( iter < itermax ) and
            ( fgam >= minstep )
            ):        

            c_next = c - fgam * gam * grad # update
            c_next[:,0], c0_proj = _project_c0(c_next[:,0])

            loss_next = fun( c_next )
            if loss_next < loss: # check if loss function decreased
                iter += 1
                # accept new values
                c = c_next
                loss = loss_next
                # get new gradient
                grad = jac( c )
                # get precision of fit
                nGrad_m1 = nGrad
                nGrad = _vnorm(grad)
                prec = np.abs(nGrad_m1-nGrad)/nGrad
                # write into opt dictionary
                opt['loss'].append( loss )
                opt['nGrad'].append( nGrad )
                opt['proj'].append( c0_proj )
                opt['max_memory'] = max( getrusage(RUSAGE_SELF), opt['max_memory'] )

                t_iter_av = (time()-t1)/iter
                sys.stdout.write('\r\tIt %d, loss: %5.3f, t/it: %.2f s        ' % (
                    iter,loss,t_iter_av))
                sys.stdout.flush()

            else: # else reduce stepsize and try again
                fgam = fgam*0.8
                # print( '\n Reduced stepsize, current factor: %.3e' %fgam )
    finally:
        opt['c'] = c # optimized parameters
        opt['t_iter_av'] = t_iter_av # average time per iteration in s
        opt['t_total'] = time()-t0 # total optimization time in s
        opt['prec'] = prec # final precision
        opt['fgam'] = fgam # final stepsize modification
        print('\n\tFinished. time: %.2f min, precision: %.5e' % ( (time()-t0)/60, prec ))

        # write solution to file
        if name != '':
            name = '_'+name
        dt = str(datetime.now()
                ).replace(' ','_').replace('-','_').replace(':','h')[:-10]
        resname = 'opt_%s_%s_nmax%s%s_%s.txt' % (
            fit.title, modename, fit.ns[-1], name, dt)
        os.makedirs('output/', exist_ok=True)
        with open('output/'+resname, 'wb') as file:
            dump(opt, file)
        return c, opt

def _vnorm( v ):
    " calculates the vector norm "
    if isinstance(v,np.ndarray):
        return (v**2).sum()**(1/2)
    else:
        return 1

@njit 
def _project_c0( c0 ):
    " projects any negative values to 0 "
    proj = np.any(c0<0)
    if proj:
        c0[c0<0] = 0
    return c0, proj
