import os
import numpy as np
import plotly.graph_objects as go
# from interpolation.splines import eval_linear
from scipy.spatial.transform import Rotation as sprot
import scipy.stats as stats
import matplotlib.pyplot as plt
from matplotlib import cm
import matplotlib.colors as colors
import warnings
import time

from src.misc import idx_nearest
from src.model import model
from src.data import data
from src.fitting import fitting

############################################################################################
############################################################################################

def onion( Odf, OTP, title='', save=False ):
    """ Plots the ODF in 3D quaternion representation """

    ## make a plotly figure with the full odf in one plot
    # Input:
    #   Odf: orientation distribution function, 1D numpy array
    #   OTP: 3 columns of the same shape as Odf each, containing the corresponding angles (Ome,Tta,Phi)
    # optional:
    #   Nome: desired number of points in Omega
    #   output: if true saves the figure in the output folder (overwrites any file called 'odf.html')
    #   scaleOm: multiplies the radius by sin(ome/2)  
    #
    # Shows figure in internet browser (check popup blocker)

    # calculate coordinates of the rotation points
    r = np.round(np.abs(np.sin(OTP[:,0]/2)),6)+0.00001
    x = r * np.sin(OTP[:,1]) * np.cos(OTP[:,2])
    y = r * np.sin(OTP[:,1]) * np.sin(OTP[:,2])
    z = r * np.cos(OTP[:,1])

    # Create figure
    fly = go.Figure()

    # Add traces with the data plots, one for each slider step
    fly.add_trace(
        go.Scatter3d(
            x=x, y=y, z=z,
            mode='markers',
            marker=dict(
                size=3,
                color=Odf, # set color to an array/list of desired values
                colorscale='Viridis',   # choose a colorscale
                opacity=1,
                colorbar=dict(
                    thickness=20,
                    title = "Probability density"
                    )
                ),
            ),
        )

    fly.update_layout(
        title=title,
        scene = dict(
            xaxis = {
                'showgrid': False, # thin lines in the background
                'zeroline': False, # thick line at x=0
                'visible': False,  # numbers below
            },
            yaxis = {
                'showgrid': False, # thin lines in the background
                'zeroline': False, # thick line at x=0
                'visible': False,  # numbers below
            },
            zaxis = {
                'showgrid': False, # thin lines in the background
                'zeroline': False, # thick line at x=0
                'visible': False,  # numbers below
            },
            aspectmode = 'cube',
        ),
        
        )

    # write into file and show plot
    if save:
        save_plotly( fly, title )
    fly.show()

    print("Plotted onion: %s (check web browser)" % title )

############################################################################################
############################################################################################
# Other plots

def beam(x,y,z, Beam, ib, tomo=None,
         output=False):
    
    start = time.time()

    iend = np.searchsorted(ib,2**32-1) # for sparsearray
    if iend == 0:
        print('No beam touching a relevant voxel')
        return
    
    if tomo.shape == x.shape:
        Isample = tomo/tomo.max()*0.2
        cutoff = np.where(tomo>tomo.max()*0.05)[0]
    else:
        Isample = np.ones_like(x)*0.02


    Ibeam = Beam/Beam.max()/3+0.2
    # cly = ['rgba(0, 0, 0.5, {0:f})'.format(a) for a in voxels] this was used mask inactive voxels
    cly = ['rgba(0, 0, 0.5, {0:f})'.format(a) for a in Isample[cutoff]]
    cly2 = ['rgba(0.5, 0, 0, {0:f})'.format(a) for a in Ibeam[:iend]] # intensity is encoded in the opacity of the points
    fly = go.Figure(
        data=[
        go.Scatter3d(x=x[cutoff], y=y[cutoff], z=z[cutoff], mode='markers', 
                    marker_color=cly, 
                    # opacity=0.02,
                    marker_symbol='square'),
        go.Scatter3d(x=x[ib[:iend]], y=y[ib[:iend]], z=z[ib[:iend]], 
                    mode='markers', marker_color=cly2, 
                    marker_size = 10,
                    marker_symbol='circle')
        ]
    )
    fly.show()

    if output:
        save_plotly(fly,'beam')

    end = time.time()
    print("Plotted beam, %.3f s" % (end - start) )

############################################################################################

def tomogram( mod, tomogram, scal=0.01, cut=0.05 ):

    tomo = tomogram.flatten() /tomogram.max()
    tomo[tomo<0] = 0

    mask = np.where(tomo > cut)[0]
    X,Y,Z = mod.x_p[mask,0], mod.x_p[mask,1], mod.x_p[mask,2]
    Intensity = tomo[mask] * scal
    print('\tMaking tomogram, No of points: %d' % Intensity.shape[0])
    cly = ['rgba(0.5, 0, 0, {0:f})'.format(a) for a in Intensity] # intensity is encoded in the opacity of the points
    fly = go.Figure(
        data=[
        go.Scatter3d(x=X, y=Y, z=Z, 
                    mode='markers', marker_color=cly, 
                    marker_size = 10,
                    marker_symbol='circle')
        ]
    )
    fly.show()

############################################################################################

def singlefit2D( mod:model, fit:fitting, g, t, C,
        polar = False, 
        log=True, show_off=False,
        output=False, show=False ):
    
    shape = mod.detShape
    PHi = mod.Chi_det.reshape(shape)
    QQ = mod.Qq_det.reshape(shape)

    im = fit.image(C,g,t)
    Fit = np.zeros( shape[0]*shape[1], np.float64)
    Fit[fit.mask_detector] = im
    FIt = Fit.reshape(shape) #* fit.sim_fac # reshape and scale

    Data = np.zeros_like(Fit)
    Data[fit.mask_detector] = fit.data[g,t]
    DAta = Data.reshape(shape)

    maxval = max( Data.max(), FIt.max() )
    # maxval = FIt.max()

    if polar:
        f,ax = plt.subplots(1,3,figsize=(8,3),subplot_kw=dict(projection='polar'))
    else:
        f,ax = plt.subplots(1,3,figsize=(8,3),sharey=True)
        PHi = PHi*180/np.pi
    lognrm = colors.LogNorm(vmin=maxval*1e-2, vmax=maxval)
    if log:
        nrm = colors.LogNorm(vmin=maxval*1e-2, vmax=maxval)
        res = abs(DAta-FIt)
        ax[2].set_title('abs(Residuals)')
    else:
        nrm = colors.Normalize(vmin=-maxval, vmax=maxval)
        res = DAta - FIt
        ax[2].set_title('Data-Fit')
    if show_off:
        DAta[DAta<=0] = maxval*1e-2
        FIt[FIt<=0] = maxval*1e-2
    ax[0].pcolormesh(PHi, QQ, DAta, norm=lognrm, cmap='plasma')
    m= ax[1].pcolormesh(PHi, QQ, FIt, norm=lognrm, cmap='plasma')
    m=ax[2].pcolormesh(PHi, QQ, res, norm=nrm, cmap='RdBu_r' )
    ax[0].set_title('Data')
    ax[1].set_title('Fit')
    ax[0].set_xticks([0, 90, 180, 270])
    ax[1].set_xticks([0, 90, 180, 270])
    ax[0].set_xlabel('$\chi$')
    ax[1].set_xlabel('$\chi$')
    ax[0].set_ylabel('q [nm$^{-1}$]')
    f.colorbar(m, label='Scattering Intensity', ax=ax[2])
    f.tight_layout()

    if output:
        if not os.path.exists('output'):
            os.makedirs('output')
        f.savefig('output/fitimage.pdf')
    
    if show:
        plt.show()

############################################################################################


def save_plotly( figure, filename ):
    """ Saves html image to output folder """
    if not os.path.exists('output'):
        os.makedirs('output')
    path = os.path.join( 'output', filename+'.html')
    figure.write_html( path )
    print('\tSaved figure to: %s' % path )